import shlex
import subprocess
import os
import zipfile
import shutil
import logging
import sys

logging.basicConfig(level=logging.INFO)

def build_darknet(darknet_dir, branch_name, target_location):
    '''
    Utility method to download and install darknet
    :param download_path:
    :return:
    '''
    import requests
    download_path = darknet_dir
    logging.info("Temp Path: "+ download_path)

    logging.info("Downloading darknet")
    os.makedirs(download_path, exist_ok=True)
    response = requests.get('https://github.com/madhawav/darknet/archive/'+branch_name+'.zip')

    logging.info("Extracting darknet")
    with open(os.path.join(download_path,"darknet.zip"), "wb") as f:
        f.write(response.content)

    zip_ref = zipfile.ZipFile(os.path.join(download_path,"darknet.zip"), 'r')
    zip_ref.extractall(download_path)
    zip_ref.close()

    os.remove(os.path.join(download_path, "darknet.zip"))

    logging.info("Building darknet")
    build_ret = subprocess.Popen("make", shell=True, stdout=subprocess.PIPE, cwd=os.path.join(download_path,"darknet-"+branch_name))

    for line in iter(build_ret.stdout.readline, ''):
        if len(line) != 0:
            logging.info(line.rstrip())
        else:
            break

    if build_ret.wait() == 0:
        logging.info("Darknet building successful")
    else:
        return False

    logging.info("Moving to " + target_location)
    shutil.move(os.path.join(download_path,"darknet-"+branch_name+"/libdarknet.so"), target_location)

    return True

def clean_darknet(darknet_path):
    '''
    Cleanup darknet download
    :param darknet_path:
    :return:
    '''
    shutil.rmtree(darknet_path,ignore_errors=True)

def get_python_cflags():
    '''
    Utility method to retrieve compiler flags to compile a python library. Uses 'python3-config --cflags'.
    '''
    command = "python3-config --cflags"
    proc = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()

    return out.rstrip().decode('utf-8').split()

def get_python_libs():
    '''
    Utility method to retrieve linker flags to build a python library. Uses 'python3-config --libs'.
    '''
    command = "python3-config --libs"
    proc = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()    
    
    return out.rstrip().decode('utf-8').split()

def get_readme():
    '''
    Retrieve readme of the package.
    '''
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),"pypi_readme.md"),"r") as f:
        return f.read()

# Code based on https://github.com/matze/pkgconfig
def get_cflags(package):
    call_name = "pkg-config"
    if 'PKG_CONFIG' in os.environ:
        call_name = os.environ['PKG_CONFIG']

    command = call_name + " --cflags " + package

    proc = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()

    response = out.rstrip().decode('utf-8').split()
    if len(response) == 0:
        return None
    return response

def find_site_packages():
    site_packages = [p for p in sys.path if p.endswith("site-packages") or p.endswith("site-packages/")]
    return site_packages

def find_dist_packages():
    dist_packages = [p for p in sys.path if p.endswith("dist-packages") or p.endswith("dist-packages/")]
    return dist_packages

def get_libs(package):
    call_name = "pkg-config"
    if 'PKG_CONFIG' in os.environ:
        call_name = os.environ["PKG_CONFIG"]

    command = call_name + " --libs " + package

    proc = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()

    response = out.rstrip().decode('utf-8').split()
    if len(response) == 0:
        return None
    return response


# distutils: language = "c++"

import numpy as np
cimport numpy as np

cdef extern from "bridge.h":
    cdef int CV_8UC3

    cdef cppclass Mat:
        Mat() except +
        void create(int, int, int)
        void* data
        int rows
        int cols
        int channels()
        void deallocate()
        void release()

    ctypedef struct box:
        float x, y, w, h

    ctypedef struct detection:
        box bbox
        int classes
        float *prob
        float *mask
        float objectness
        int sort_class

    cdef struct network:
        pass

    ctypedef struct metadata:
        int classes
        char** names

    ctypedef struct image:
        int w
        int h
        int c
        float *data

    float *network_predict_image(network *net, image im)

    int getCV_8UC3()
    image get_darknet_image(const Mat &input)

    network* load_network(char* config, char* weights, int p)

    metadata get_metadata(char *file)

    void free_network(network* net)

    void free_detections(detection *dets, int n)

    void do_nms_obj(detection *dets, int total, int classes, float thresh)

    void show_image(image img, char* title, int ms)
    void free_image(image img)

    detection *get_network_boxes(network *net, int w, int h, float thresh, float hier, int *map, int relative, int *num)

    IF USE_GPU == 1:
        void cuda_set_device(int n)



import tempfile
import os.path as osp
from distutils.command.build import build
from distutils.command.clean import clean
import sys
import numpy as np # TODO: Need a mechanism to ensure numpy is already installed. Define a pyproject.toml file.
import shutil

# Compile using .cpp files if cython is not present
try:
    from Cython.Distutils import build_ext
except ImportError:
    from distutils.command.build_ext import build_ext
    use_cython = False
else:
    use_cython = True

from setuptools import setup, Extension
from setup_util import build_darknet, clean_darknet, get_cflags, get_libs, find_site_packages, get_readme, find_dist_packages, get_python_libs, get_python_cflags
import logging
import os

logging.basicConfig(level=logging.INFO)

# Default configuration
USE_GPU = False
USE_CV = False
DEFAULT_CUDA_HOME = "/usr/local/cuda"

if "GPU" in os.environ:
    if "DARKNET_HOME" in os.environ:
        logging.warning("GPU environment variable is skipped since DARKNET_HOME is specified")
        if int(os.environ["GPU"]) == 1:
            USE_GPU = True
        else:
            USE_GPU = False
    else:
        if int(os.environ["GPU"]) == 1:
            logging.info("Darknet will be compiled with GPU support")
            USE_GPU = True
        else:
            logging.info("Darknet will be compiled without GPU support")
            USE_GPU = False


if "OPENCV" in os.environ and int(os.environ["OPENCV"]) == 0:
    logging.info("Compiling wrapper without OpenCV")
    USE_CV = False
elif "OPENCV" in os.environ and int(os.environ["OPENCV"]) == 1:
    logging.info("Compiling wrapper with OpenCV")
    USE_CV = True

if USE_CV & (get_libs("opencv") is None or get_cflags("opencv") is None):
    logging.warning("OpenCV is not configured. Compiling wrapper without OpenCV!")
    USE_CV = False


if USE_GPU:
    if USE_CV:
        build_branch_name = "yolo34py-intergration-v2"
    else:
        build_branch_name = "yolo34py-intergration-nocv-v2"
else:
    build_branch_name = "yolo34py-intergration-nogpu-v2"
    if "DARKNET_HOME" not in os.environ:
        if USE_CV:
            logging.warning("Non GPU darknet branch is used. Compiling wrapper without OpenCV!")
        USE_CV = False # OpenCV requires yolo34py-intergration branch which has OpenCV enabled

if "DARKNET_HOME" not in os.environ:
    logging.info("Selected Darknet Branch: " + build_branch_name+ " from Darknet Fork 'https://github.com/madhawav/darknet/'")


temp_dir =  os.path.join(tempfile.gettempdir(), "darknet") # Temp directory to build darknet

# Check whether user has specified DARKNET_HOME directory. If so, we would use the darknet installation at this location.
if not "DARKNET_HOME" in os.environ:
    darknet_dir = os.path.join(temp_dir, "darknet-" + build_branch_name)
else:
    logging.info("DARKNET_HOME is set: " + os.environ["DARKNET_HOME"])
    darknet_dir = os.environ["DARKNET_HOME"]

include_paths = [np.get_include(), os.path.join(darknet_dir,"include"), os.path.join(darknet_dir,"src")]
libraries = ["darknet","m", "pthread"]
library_paths = [".", "./__libdarknet"]

extra_compiler_flags = get_python_cflags()
extra_linker_flags = get_python_libs()

cython_compile_directives = {}
macros = []

if USE_GPU:
    cuda_home = None
    if "CUDA_HOME" in os.environ:
        cuda_home = os.environ["CUDA_HOME"]
    elif osp.exists(DEFAULT_CUDA_HOME):
        cuda_home = DEFAULT_CUDA_HOME

    if cuda_home is not None:
        include_paths.append(os.path.join(cuda_home,"include"))
    else:
        raise Exception("Environment variable CUDA_HOME not set")
    cython_compile_directives["USE_GPU"] = 1
    macros.append(("USE_GPU", 1))
else:
    cython_compile_directives["USE_GPU"] = 0
    macros.append(("USE_GPU", 0))

if USE_CV:
    extra_compiler_flags.extend(get_cflags("opencv"))
    extra_linker_flags.extend(get_libs("opencv"))
    cython_compile_directives["USE_CV"] = 1
    macros.append(("USE_CV", 1))
else:
    cython_compile_directives["USE_CV"] = 0
    macros.append(("USE_CV", 0))


# Add linker flag to search in site_packages/__libdarknet. libdarknet.so is located at this location.
for site_package in find_site_packages():
    extra_linker_flags.append("-Wl,-rpath," + os.path.join(site_package,"__libdarknet"))

for dist_package in find_dist_packages():
    extra_linker_flags.append("-Wl,-rpath," + os.path.join(dist_package,"__libdarknet"))

if "--inplace" in sys.argv:
    extra_linker_flags.append("-Wl,-rpath,.")  # Added to make test code work

if use_cython:
    pydarknet_extension = Extension("pydarknet", ["pydarknet.pyx", "pydarknet.pxd", "bridge.cpp"], include_dirs=include_paths, language="c++",
                  libraries=libraries, library_dirs=library_paths,   extra_link_args=extra_linker_flags,
                  extra_compile_args=extra_compiler_flags, define_macros = macros)

    # Pass macros to Cython
    pydarknet_extension.cython_compile_time_env = cython_compile_directives
else:
    pydarknet_extension = Extension("pydarknet", ["pydarknet.cpp", "bridge.cpp"],
                                    include_dirs=include_paths, language="c++",
                                    libraries=libraries, library_dirs=library_paths, extra_link_args=extra_linker_flags,
                                    extra_compile_args=extra_compiler_flags, define_macros=macros)

    # NOTE: It is assumed that pydarknet.cpp is already generated using pydarknet.py. It is also assumed that USE_CV
    # flag is unchanged between cythonize and current compilation.

ext_modules=[
    pydarknet_extension
]

darknet_setup_done = False

def setup_darknet():
    '''
    Configures darknet on which the wrapper works
    :return:
    '''
    global darknet_setup_done
    if darknet_setup_done:
        return

    target_location = os.path.join(os.path.dirname(os.path.abspath(__file__)), "__libdarknet", "libdarknet.so")

    if "--inplace" in sys.argv:
        logging.info("For inplace compilations, target location is set to root")
        target_location =  os.path.join(os.path.dirname(os.path.abspath(__file__)), "libdarknet.so")

    if "DARKNET_HOME" not in os.environ:
        # If user has not specified DARKNET_HOME, we will download and build darknet.
        build_darknet(temp_dir, build_branch_name, target_location)
    else:
        logging.info("Copying libdarknet.so from " + os.environ["DARKNET_HOME"])
        # If user has set DARKNET_HOME, it is assumed that he has built darknet. We will copy libdarknet.so from users location to site-pacakges/__libdarknet
        shutil.copyfile(os.path.join(os.environ["DARKNET_HOME"], "libdarknet.so"),
                        target_location)

    darknet_setup_done = True

class CustomBuild(build):
    def run(self):
        # This is triggered when src distribution is made. Not triggered for build_ext.
        setup_darknet()
        build.run(self)

class CustomBuildExt(build_ext):
    def run(self):
        setup_darknet()
        build_ext.run(self)

        if not "DARKNET_HOME" in os.environ:
            clean_darknet(temp_dir)

class CustomClean(clean):
    def run(self):
        if os.path.exists(os.path.join(os.path.dirname(os.path.abspath(__file__)),"__libdarknet","libdarknet.so")):
            logging.info("removing __libdarknet/libdarknet.so")
            os.remove(os.path.join(os.path.dirname(__file__),"__libdarknet","libdarknet.so"))

        if os.path.exists(os.path.join(os.path.dirname(os.path.abspath(__file__)), "libdarknet.so")):
            logging.info("removing libdarknet.so")
            os.remove(os.path.join(os.path.dirname(os.path.abspath(__file__)),"libdarknet.so"))

        if os.path.exists(os.path.join(os.path.dirname(os.path.abspath(__file__)),"pydarknet.cpp")):
            logging.info("removing pydarknet.cpp")
            os.remove(os.path.join(os.path.dirname(os.path.abspath(__file__)),"pydarknet.cpp"))

        for f in os.listdir(os.path.dirname(os.path.abspath(__file__))):
            if f.startswith("pydarknet.") and f.endswith(".so"):
                logging.info("removing " + f)
                os.remove(os.path.join(os.path.dirname(os.path.abspath(__file__)),f))

        clean.run(self)


if USE_GPU:
    name = "yolo34py-gpu"
else:
    name = "yolo34py"

cmd_class = {'clean': CustomClean, "build": CustomBuild, "build_ext": CustomBuildExt}


setup(
  name = name,
  description="Python wrapper on YOLO 3.0 implementation by 'pjreddie': (https://pjreddie.com/yolo)",
  long_description=get_readme(),
  long_description_content_type="text/markdown",
  cmdclass= cmd_class,
  version='0.1.rc61',
  ext_modules = ext_modules,
  platforms=["linux-x86_64"],
  setup_requires=[
      'cython>=0.27',
      'requests',
      'numpy'
  ],
  install_requires=[
      'cython>=0.27',
      'requests',
      'numpy'
  ],
  zip_safe=False,
  python_requires='>=3.5',
  author='Madhawa Vidanapathirana',
  author_email='madhawavidanapathirana@gmail.com',
  url="https://github.com/madhawav/YOLO3-4-Py",
  package_dir={"__libdarknet": "__libdarknet"},
  packages=["__libdarknet"],
  include_package_data=True,
  license="YOLO34Py wrapper is under Apache 2.0. Darknet is Public Domain.",
  classifiers=[
      'Development Status :: 4 - Beta',
      'License :: OSI Approved :: Apache Software License',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',
      'Topic :: Text Processing :: Linguistic',
      'Operating System :: POSIX :: Linux',
      'Intended Audience :: Science/Research',
      'Topic :: Scientific/Engineering :: Artificial Intelligence'
  ],
  keywords="yolo darknet object detection vision",

)



# distutils: language = "c++"

import numpy as np

from libc.string cimport memcpy
from libc.stdlib cimport malloc

is_compiled_with_gpu = bool(USE_GPU)
is_compiled_with_opencv = bool(USE_CV)

IF USE_GPU == 1:
    def set_cuda_device(n):
        cuda_set_device(n)

cdef class Image:
    cdef image img;

    def __cinit__(self, np.ndarray ary):
        """
        Initialize a Darknet Image using a Numpy Array. Provide the input in BGR format.
        :param ary: Image in BGR order.
        :return:
        """
        IF USE_CV == 1:
            # Code adapted from https://github.com/solivr/cython_opencvMat
            assert ary.ndim==3 and ary.shape[2]==3, "ASSERT::3channel RGB only!!"

            cdef np.ndarray[np.uint8_t, ndim=3, mode ='c'] np_buff = np.ascontiguousarray(ary, dtype=np.uint8)
            cdef unsigned int* im_buff = <unsigned int*> np_buff.data
            cdef int r = ary.shape[0]
            cdef int c = ary.shape[1]
            cdef Mat m
            m.create(r, c, CV_8UC3)
            memcpy(m.data, im_buff, r*c*3)
            # End of adapted code block

            self.img = get_darknet_image(m)
            m.release()
        ELSE:
             # Code adapted from https://github.com/solivr/cython_opencvMat
            assert ary.ndim==3 and ary.shape[2]==3, "ASSERT::3channel RGB only!!"

            # Re-arrange to suite Darknet input format
            ary = ary.transpose(2, 0, 1)

            # BGR to RGB
            ary = ary[::-1,:,:]

            # 0..1 Range
            ary = ary/255.0

            # To c_array
            cdef np.ndarray[np.float32_t, ndim=3, mode ='c'] np_buff = np.ascontiguousarray(ary, dtype=np.float32)
            cdef int c = ary.shape[0]
            cdef int h = ary.shape[1]
            cdef int w = ary.shape[2]

            # Copy to Darknet image
            self.img.w = w
            self.img.h = h
            self.img.c = c
            self.img.data = <float*>malloc(h*w*c*4)
            memcpy(self.img.data, np_buff.data, h*w*c*4)


    def show_image(self, char* title, int wait_duration_in_ms = 1):
        """
        Display image in a window/
        :param title: Title of window
        :param wait_duration_in_ms: Wait duration to block
        :return:
        """
        show_image(self.img, title, wait_duration_in_ms)

    def __dealloc__(self):
        free_image(self.img)

cdef class Detector:
    cdef network* net
    cdef metadata meta

    def __cinit__(self, char* config, char* weights, int p, char* meta):
        """
        Initialize a Darknet Model.
        :param config: Path to config file.
        :param weights: Path to weights file.
        :param p: Pass Zero
        :param meta: Path to coco.data file.
        :return:
        """
        self.net = load_network(config, weights, p)
        self.meta = get_metadata(meta)

    # Code adapted from https://github.com/pjreddie/darknet/blob/master/python/darknet.py

    def classify(self, Image img):
        """
        Classify an image using the model
        :param img: Image to be classified
        :return: Sorted list of <Label ID, Score> tuples.
        """
        out = network_predict_image(self.net, img.img)
        res = []
        for i in range(self.meta.classes):
            res.append((self.meta.names[i], out[i]))
        res = sorted(res, key=lambda x: -x[1])
        return res

    def detect(self, Image image, float thresh=.5, float hier_thresh=.5, float nms=.45):
        """
        Detect objects in an image using the model.
        :param image: Image to process.
        :param thresh: Threshold parameter.
        :param hier_thresh: Hier Threshold Parameter.
        :param nms: None maximal suppression parameter.
        :return:
        """
        cdef int num = 0
        cdef int* pnum = &num
        network_predict_image(self.net, image.img)
        dets = get_network_boxes(self.net, image.img.w, image.img.h, thresh, hier_thresh, <int*>0, 0, pnum)

        num = pnum[0]
        if (nms > 0):
            do_nms_obj(dets, num, self.meta.classes, nms)

        res = []
        for j in range(num):
            for i in range(self.meta.classes):
                if dets[j].prob[i] > 0:
                    b = dets[j].bbox
                    res.append((self.meta.names[i], dets[j].prob[i], (b.x, b.y, b.w, b.h)))
        res = sorted(res, key=lambda x: -x[1])

        free_detections(dets, num)
        return res

    # End of adapted code block

    def __dealloc__(self):
        free_network(self.net)